import improveai.decision_model as dm


class GivensProvider:

    # TODO how givens should be provided - should it be some sort of customized
    #  implementation for each model dynamically providing givens?
    #  also in iOS this method returns @{} and is named givensForModel
    #  (but maybe I got this wrong)
    #  I made it return provided givens -> which implementation is correct
    def givens(self, for_model, givens: dict or None = {}) -> dict:
        """
        Passthrough method

        Parameters
        ----------
        for_model: dm.DecisionModel
            instance of DecisionModel for which givens should be provided
        givens: dict or None
            givens which will be returned

        Returns
        -------
        dict or None
            givens which will be returned

        """
        assert isinstance(for_model, dm.DecisionModel)
        assert isinstance(givens, dict) or givens is None
        return givens



#!python3
#cython: language_level=3

cdef extern from "npy_no_deprecated_api.h": pass

import cython
from libc.math cimport isnan
import numpy as np
cimport numpy as np
import xxhash

cpdef object xxhash3 = xxhash.xxh3_64_intdigest

# from improveai.cythonized_feature_encoding.cythonized_feature_encoding_utils \
#     import encoded_variant_into_np_row, encode_variants_multiple_givens, \
#     encoded_variants_to_np
import improveai.cythonized_feature_encoding.cythonized_feature_encoding_utils as cfeu


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef encode(object object_, unsigned long long seed, double small_noise, dict features):

    cdef str feature_name = None
    cdef double previous_object_

    if isinstance(object_, (int, float)):  # bool is an instanceof int
        if isnan(object_):  # nan is treated as missing feature, return
            return

        feature_name = hash_to_feature_name(seed)

        previous_object_ = \
            _get_previous_value(
                feature_name=feature_name, into=features,
                small_noise=small_noise)

        features[feature_name] = \
            sprinkle(object_ + previous_object_, small_noise)

        return

    cdef unsigned long long hashed
    cdef double previous_hashed
    cdef str hashed_feature_name
    cdef double previous_hashed_for_feature_name

    if isinstance(object_, str):
        hashed = xxhash3(object_, seed=seed)

        feature_name = hash_to_feature_name(seed)

        previous_hashed = \
            _get_previous_value(
                feature_name=feature_name, into=features,
                small_noise=small_noise)

        features[feature_name] = \
            sprinkle(
                <double>(<long long>(((hashed & 0xffff0000) >> 16) - 0x8000)) + previous_hashed,
                small_noise)

        hashed_feature_name = hash_to_feature_name(hashed)

        previous_hashed_for_feature_name = \
            _get_previous_value(
                feature_name=hashed_feature_name, into=features,
                small_noise=small_noise)

        features[hashed_feature_name] = \
            sprinkle(
                <double>(<long long>((hashed & 0xffff) - 0x8000)) + previous_hashed_for_feature_name,
                small_noise)

        return

    if isinstance(object_, dict):
        for key, value in object_.items():
            encode(
                value, xxhash3(key, seed=seed), small_noise, features)
        return

    if isinstance(object_, (list, tuple)):
        for index, item in enumerate(object_):
            encode(
                item, xxhash3(index.to_bytes(8, byteorder='big'), seed=seed),
                small_noise, features)

        return

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double _get_previous_value(str feature_name, dict into, double small_noise):
    """
    Gets previous, 'unsprinkled' value of <feature_name> feature
    from <features> dict

    Parameters
    ----------
    feature_name: str
        name of a feature which previous value is desired
    into: dict
        dict containing current results of encoding
    small_noise: float
        small noise used to sprinkle

    Returns
    -------
    float
        'unsprinkled' value of desired feature

    """

    cdef double  previous_sprinkled_object_

    if into.get(feature_name, None) is None:
        return 0.0
    else:
        previous_sprinkled_object_ = into.get(feature_name, None)
        return reverse_sprinkle(previous_sprinkled_object_, small_noise)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef str hash_to_feature_name(unsigned long long hash_):
    return '%0*x' % (8, (hash_ >> 32))


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double shrink(double noise):
    return noise * pow(2, -17)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double sprinkle(double x, double small_noise):
    return (x + small_noise) * (1 + small_noise)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double reverse_sprinkle(double sprinkled_x, double small_noise):
    """
    Retrieves true value from sprinkled one

    Parameters
    ----------
    sprinkled_x: float
        sprinkled value
    small_noise: float
        noise used to calculate sprinkled_Xx

    Returns
    -------
    float
        true value of x which was used to calculate sprinkled_x

    """
    return sprinkled_x / (1 + small_noise) - small_noise


cdef class FeatureEncoder:
    """
    This class serves as a preprocessor which allows to
    convert event data in form of a JSON line (variants, givens and
    extra features) to the tabular format expected on input by XGBoost.

    Short summary of methods:
     - encode_givens() -> encodes givens / variant metadata to dict of
       feature name -> float pairs

     - encode_variant() -> encodes variant to dict of feature name -> float pairs

     - encode_feature_vector() -> encodes provided variant and givens;
       appends extra features to the encoded variant dict;
       converts encoded variant to numpy array using provided feature names

     - add_noise() -> sprinkles each value of the input dict with small noise

     - encode_to_np_matrix()  -> encodes collection of variants with collection
       of givens and extra features to the numpy 2D array / matrix. If desired
       uses cython backed (works much faster)

     - add_extra_features() - helper method for adding extra features to already
       encoded variants

    """

    cdef unsigned long long variant_seed
    cdef unsigned long long value_seed
    cdef unsigned long long givens_seed

    def __init__(self, long long model_seed):
        if (model_seed < 0):
            raise TypeError(
                "xxhash3_64 takes an unsigned 64-bit integer as seed. "
                "Seed must be greater than or equal to 0.")

        # memoize commonly used seeds
        self.variant_seed = xxhash3("variant", seed=model_seed)
        self.value_seed = xxhash3("$value", seed=self.variant_seed)
        self.givens_seed = xxhash3("givens", seed=model_seed)

    cdef void _check_noise_value(self, double noise) except *:
        if noise > 1.0 or noise < 0.0:
            raise ValueError(
                'Provided `noise` is out of allowed bounds <0.0, 1.0>')

    cpdef dict encode_givens(self, dict givens, double noise=0.0, dict into=None):

        self._check_noise_value(noise=noise)

        if givens is not None and not isinstance(givens, dict):
            raise TypeError(
                "Only dict type is supported for context encoding. {} type was "
                "provided.".format(type(givens)))

        if into is None:
            into = {}

        encode(givens, self.givens_seed, shrink(noise), into)

        # TODO wait until the conversion mechanism is determined
        # self._convert_values_to_float32(into=into)
        return into

    cpdef dict encode_variant(self, object variant, double noise=0.0, dict into=None):

        # TODO maybe this check is only a waste of runtime
        self._check_noise_value(noise=noise)

        if into is None:
            into = {}

        small_noise = shrink(noise)

        if isinstance(variant, dict):
            encode(variant, self.variant_seed, small_noise, into)
        else:
            encode(variant, self.value_seed, small_noise, into)

        # TODO wait until the conversion mechanism is determined
        # self._convert_values_to_float32(into=into)
        return into


    cpdef encode_feature_vector(
            self, object variant = None, dict givens = None,
            dict extra_features = None, list feature_names = None,
            double noise = 0.0, np.ndarray into = None):

        if into is None:
            raise ValueError('`into` can`t be None')

        cdef dict encoded_givens = {}
        # encode givens
        encoded_givens = \
            self.encode_givens(givens=givens, noise=noise, into=dict())

        cdef dict encoded_variant = {}
        # encoded variant and givens
        encoded_variant = \
            self.encode_variant(
                variant=variant, noise=noise, into=encoded_givens)

        if extra_features:
            if not isinstance(extra_features, dict):
                raise TypeError(
                    'Provided `extra_features` is not a dict but: {}'
                    .format(extra_features))

            encoded_variant.update(extra_features)

        # n + nan = nan so you'll have to check for nan values on into

        # i7 10th gen time per iter - 7.464-05 [s] / 0.96% of pure python time
        cfeu.encoded_variant_into_np_row(
            encoded_variant=encoded_variant, feature_names=feature_names,
            into=into)

    cpdef double [:, :] encode_to_np_matrix(
            self, list variants, list multiple_givens,
            list multiple_extra_features, list feature_names, double noise):
        """
        Provided list of variants and corresponding lists of givens and extra
        features encodes variants completely and converts to numpy array

        Parameters
        ----------
        variants: list
            list of variants to be encoded
        multiple_givens: list
            list of givens - givens[i] will be used to encode variants[i]
        multiple_extra_features: list
            list of extra_features - multiple_extra_features[i] will be used
            to encode variants[i]
        feature_names: list
            names of features expected by the model
        noise: float
            noise to be used when encoding data

        Returns
        -------
        np.ndarray
            2D numpy array of encoded variants

        """

        # i7 10th gen time per variant - 2.778e-05 [s] |
        # 33% of pure python implementation runtime per iter
        encoded_variants = cfeu.encode_variants_multiple_givens(
            variants=variants, multiple_givens=multiple_givens,
            multiple_extra_features=multiple_extra_features, noise=noise,
            feature_encoder=self.encode_variant,
            givens_encoder=self.encode_givens)
        encoded_variants_array = cfeu.encoded_variants_to_np(
            encoded_variants=encoded_variants, feature_names=feature_names)

        return encoded_variants_array

    cpdef void add_extra_features(self, list encoded_variants, list extra_features):
        """
        Once variants are encoded this method can be used to quickly append
        extra features

        Parameters
        ----------
        encoded_variants: list
            collection of encoded variants to be updated with extra features
        extra_features: list
            payload to be appended to encoded variants

        Returns
        -------
        None
            None

        """

        if extra_features is None:
            return

        if not isinstance(encoded_variants, list):
            raise TypeError('`encoded_variants` should be of a list type')

        if not isinstance(extra_features, list):
            raise TypeError('`extra_features` should be of a list type')

        [encoded_variant.update(single_extra_features)
         if single_extra_features is not None else encoded_variant
         for encoded_variant, single_extra_features in
         zip(encoded_variants, extra_features)]

    cpdef _convert_values_to_float32(self, dict into):
        """
        Converts all values in the input dict to float32

        Parameters
        ----------
        into: dict
            converted dict

        Returns
        -------

        """
        for k in into.keys():
            into[k] = np.float32(into[k])


from abc import ABC, abstractmethod
from copy import deepcopy
import numpy as np
from typing import Dict

from improveai.feature_encoder import FeatureEncoder
from improveai.utils.url_tools import is_path_http_addr, \
    get_model_bytes_from_url
from improveai.utils.gzip_tools import check_and_get_unzpd_model


class BasicChooser(ABC):
    MODEL_NAME_REGEXP = "^[a-zA-Z0-9][\w\-.]{0,63}$"

    @property
    @abstractmethod
    def model(self) -> object:
        pass

    @model.setter
    @abstractmethod
    def model(self, new_val: object):
        pass

    @property
    @abstractmethod
    def model_metadata_key(self):
        pass

    @model_metadata_key.setter
    @abstractmethod
    def model_metadata_key(self, new_val: str):
        pass

    @property
    @abstractmethod
    def feature_encoder(self) -> FeatureEncoder:
        pass

    @feature_encoder.setter
    @abstractmethod
    def feature_encoder(self, new_val: FeatureEncoder):
        pass

    @property
    @abstractmethod
    def model_metadata(self) -> Dict[str, object]:
        pass

    @model_metadata.setter
    @abstractmethod
    def model_metadata(self, new_val: Dict[str, object]):
        pass

    @property
    @abstractmethod
    def model_seed_key(self) -> str:
        pass

    @model_seed_key.setter
    @abstractmethod
    def model_seed_key(self, new_val: str):
        pass
    
    @property
    @abstractmethod
    def model_seed(self):
        pass 
    
    @model_seed.setter
    @abstractmethod
    def model_seed(self, value):
        pass
    
    @property
    @abstractmethod
    def model_name_key(self):
        pass
    
    @model_name_key.setter
    def model_name_key(self, value):
        pass

    @property
    @abstractmethod
    def model_name(self):
        pass

    @model_name.setter
    def model_name(self, value):
        pass

    @property
    @abstractmethod
    def model_feature_names_key(self):
        return

    @model_feature_names_key.setter
    @abstractmethod
    def model_feature_names_key(self, value):
        pass

    @property
    @abstractmethod
    def model_feature_names(self):
        return

    @model_feature_names.setter
    @abstractmethod
    def model_feature_names(self, value):
        pass

    @abstractmethod
    def load_model(self, inupt_model_src, **kwargs):
        pass

    @abstractmethod
    def score(self, variants, givens, **kwargs):
        pass

    @abstractmethod
    def _get_model_metadata(self, **kwargs):
        pass

    @staticmethod
    def get_model_src(model_src: str or bytes) -> str or bytes:
        """
        Gets model src from provided input path, url or bytes
        
        Parameters
        ----------
        model_src: str or bytes
            pth to model, url or bytes

        Returns
        -------
        str or bytes
            path or downloaded model

        """
        raw_model_src = model_src
        if is_path_http_addr(pth_to_model=model_src):
            raw_model_src = get_model_bytes_from_url(model_url=model_src)

        unzpd_model_src = check_and_get_unzpd_model(model_src=raw_model_src)
        return unzpd_model_src

    def _get_features_count(self) -> int:
        table = self.model_metadata.get(self.lookup_table_key, None)
        return len(table[1])

    def _get_model_feature_names(self, model_metadata: dict):
        """
        Getter for model features

        Returns
        -------
        dict
            dict with feature names extracted from model metadata

        """

        if not model_metadata:
            raise ValueError('Model metadata empty or None!')

        feature_names = model_metadata.get(self.model_feature_names_key, None)

        if not feature_names:
            raise ValueError('Feature names not in model metadata!')
        
        return feature_names

    def _get_model_seed(self, model_metadata: dict):
        if not model_metadata:
            raise ValueError('Model metadata empty or None!')

        model_seed = model_metadata.get(self.model_seed_key, None)

        if not model_seed:
            raise ValueError('Feature names not in model metadata!')

        return model_seed

    def _get_model_name(self, model_metadata: dict):
        if not model_metadata:
            raise ValueError('Model metadata empty or None!')

        model_name = model_metadata.get(self.model_name_key, None)

        if not model_name:
            raise ValueError('Feature names not in model metadata!')

        return model_name

    # TODO sort and choose are deprecated (kept only for CLI compatibility)
    def sort(
            self, variants_w_scores: np.ndarray,
            scores_col_idx: int = 1, class_cols_idx: int = 2) -> np.ndarray:
        """
        Performs sorting of provided variants with scores array

        Parameters
        ----------
        variants_w_scores: np.ndarray
            array with variant, scores rows
        scores_col_idx: int
            the index of column with scores
        class_cols_idx: int
            index of the class label in a single row


        Returns
        -------
        np.ndarray
            2D sorted array of rows (variant, score)

        """

        desc_scores_sorting_col = -1 * variants_w_scores[:, scores_col_idx]
        class_sorting_col = variants_w_scores[:, class_cols_idx]

        ind = np.lexsort((desc_scores_sorting_col, class_sorting_col))

        srtd_variants_w_scores = \
            variants_w_scores[ind]

        return srtd_variants_w_scores

    def choose(
            self, variants_w_scores: np.ndarray,
            scores_col_idx: int = 1, class_col_idx: int = 2) -> np.ndarray:
        """
        Chooses the variant with the highest score. Randomly breaks ties

        Parameters
        ----------
        variants_w_scores: np.ndarray
            2D array of (variant, score) rows
        scores_col_idx: int
            int indicating column with scores
        class_col_idx: int
            index of the class label in a single row

        Returns
        -------
        np.ndarray
            1D array with <best variant, best_score>

        """
        # must break ties

        choices = []

        for curr_class_id in np.unique(variants_w_scores[:, class_col_idx]):
            curr_class_variants_w_scores = \
                variants_w_scores[
                    variants_w_scores[:, class_col_idx] == curr_class_id, :]

            scores = curr_class_variants_w_scores[:, scores_col_idx]

            top_variants_w_scores = \
                curr_class_variants_w_scores[scores == scores.max()]

            chosen_top_variant_idx = \
                np.random.randint(top_variants_w_scores.shape[0])

            choices.append(top_variants_w_scores[chosen_top_variant_idx])

        return np.array(choices)



import numpy as np
import os
import pyximport

CYTHON_ENCODING_DIR_PATH = os.sep.join(__file__.split(os.sep)[:-1])

pyximport.install(setup_args={'include_dirs': [np.get_include(), CYTHON_ENCODING_DIR_PATH]})

import improveai.cythonized_feature_encoding.cythonized_feature_encoding_utils as cfeu
import improveai.cythonized_feature_encoding.cythonized_feature_encoder as cfe



import asyncio
import re
import warnings

import numpy as np

from improveai.choosers.basic_choosers import BasicChooser
from improveai.choosers.mlmodel_chooser import MLModelChooser
from improveai.choosers.xgb_chooser import NativeXGBChooser
import improveai.decision as d
import improveai.decision_context as dc
import improveai.decision_tracker as dt
import improveai.givens_provider as gp
from improveai.settings import DEBUG
from improveai.utils.general_purpose_tools import constant, check_variants, \
    get_variants_for_which


class DecisionModel:

    SUPPORTED_CALLS = ['score', 'top_scoring_variant', 'rank', 'get']
    MODEL_NAME_REGEXP = "^[a-zA-Z0-9][\w\-.]{0,63}$"

    @property
    def model_name(self) -> str:
        return self.__model_name

    @model_name.setter
    def model_name(self, value: str):

        if value is not None:
            assert isinstance(value, str)
            assert re.search(DecisionModel.MODEL_NAME_REGEXP, value) is not None
        self.__model_name = value

    @property
    def _tracker(self):
        return self.__tracker

    @property
    def chooser(self) -> BasicChooser:
        return self._chooser

    @chooser.setter
    def chooser(self, value: BasicChooser):
        self._chooser = value

    @property
    def givens_provider(self):
        return self._givens_provider

    @givens_provider.setter
    def givens_provider(self, value):
        self._givens_provider = value

    @property
    def track_url(self):
        return self._track_url

    @track_url.setter
    def track_url(self, value):
        self._track_url = value

    @property
    def id_(self):
        return self._id_

    @id_.setter
    def id_(self, value):
        self._id_ = value

    @constant
    def TIEBREAKER_MULTIPLIER() -> float:
        return 2**-23

    def __init__(
            self, model_name: str, track_url: str = None, track_api_key: str = None):
        # self.__set_model_name(model_name=model_name)
        self.model_name = model_name
        self.track_url = track_url

        self.__tracker = None
        if self.track_url:
            self.__tracker = \
                dt.DecisionTracker(track_url=self.track_url, track_api_key=track_api_key)

        self.id_ = None
        self.chooser = None
        self.givens_provider = gp.GivensProvider()

    def load(self, model_url: str):
        """
        Synchronously loads XGBoost model from provided path, creates instance
        of DecisionModel and returns it

        Parameters
        ----------
        model_url: str
            path to desired model (FS path or url)

        Returns
        -------
        DecisionModel
            new instance of decision model

        """

        self.chooser = DecisionModel._get_chooser(model_url=model_url)
        self._resolve_model_name()

        return self

    def _resolve_model_name(self):
        """
        If model name is not set loaded model name is used, otherwise model name
        from constructor `wins`

        Returns
        -------

        """
        # TODO unittest this
        if self.model_name is None:
            self.model_name = self.chooser.model_name
        else:
            self.chooser.model_name = self.model_name
            warnings.warn(
                'Model name passed to the constructor: {} will not be '
                'overwritten by loaded model name: {}.'
                .format(self.model_name, self.chooser.model_name))

    # TODO check if it should be static or not
    # TODO remove mlmodel support
    # TODO make sure what a callback should do (probably based on iOS implementation)
    @staticmethod
    def _get_chooser(
            model_url: str, loop: asyncio.BaseEventLoop = None) -> BasicChooser:
        """
        Synchronously loads XGBoost model from provided path, creates instance
        of Chooser and returns it

        Parameters
        ----------
        model_url: str
            path to desired model (FS path or url)

        Returns
        -------
        DecisionModel
            new instance of decision model

        """

        true_exception = None
        try:
            model_src = BasicChooser.get_model_src(model_src=model_url)
        except Exception as exc:
            if loop is not None:
                loop.call_exception_handler(context={
                    'message': 'Failed to load model from url: {}'.format(
                        model_url),
                    'exception': exc})
            raise exc

        chooser = None
        chooser_constructors = [MLModelChooser, NativeXGBChooser]

        for chooser_constructor in chooser_constructors:

            chooser = None

            try:
                chooser = chooser_constructor()
                chooser.load_model(input_model_src=model_src)
                break

            except Exception as exc:

                true_exception = exc

        if (chooser is None or chooser.model is None) and loop is not None:
            loop.call_exception_handler(context={
                'message': 'Failed to load model from url: {}'.format(model_url),
                'exception': true_exception})

        return chooser

    def _load_chooser_for_async(
            self, model_url: str, loop: asyncio.BaseEventLoop = None):
        """
        Wrapper around chooser loading procedure for async operation

        Parameters
        ----------
        model_url: str
            path / url to model
        loop: object
            asyncio loop

        Returns
        -------

        """

        self.chooser = DecisionModel._get_chooser(model_url=model_url, loop=loop)
        self._resolve_model_name()

    @staticmethod
    def _exception_while_loading_chooser(loop, context):
        """
        Custom exception handler for asyncio

        Parameters
        ----------
        loop: object
            asyncio loop
        context: dict
            dict with info about error

        Returns
        -------

        """
        print(
            'Model loading failed with error: {}'.format(context.get('message', None)))
        print(context.get('exception', None))

    # DEPRECATED
    # def load_async(self, model_url: str):
    #     """
    #     Loads model im an async fashion;
    #
    #     IMPORTANT:
    #     Please note that this is an EXPERIMENTAL/DEPRECATED method and might be
    #     changed in the near future
    #
    #     Parameters
    #     ----------
    #     model_url: str
    #         path / url to model
    #
    #     Returns
    #     -------
    #
    #     """
    #     loop = asyncio.get_event_loop()
    #     loop.set_exception_handler(DecisionModel._exception_while_loading_chooser)
    #     loop.run_in_executor(None, self._load_chooser_for_async, *[model_url, loop])
    #
    #     return self

    def score(self, variants: list or np.ndarray) -> np.ndarray:
        """
        Scores provided variants with available givens

        Parameters
        ----------
        variants: list or tuple or np.ndarray
            array of variants to choose from

        Returns
        -------
        np.ndarray
            array of float64 scores

        """
        # TODO make sure how givens should be created
        # in iOS the call is almost identical (nil is passed to `givens` parameter of self.givens_provider.givens())
        # TODO [CHECK] how does this correspond with DecisionContext's givens?
        givens = self.givens_provider.givens(for_model=self)
        # return equivalent of double scores
        return self._score(variants=variants, givens=givens)

    def _score(self, variants: list or np.ndarray, givens: dict) -> list or np.ndarray:
        """
        Call predict and calculate scores for provided variants

        Parameters
        ----------
        variants: list or np.ndarray
            collection of variants to be scored

        Returns
        -------
        np.ndarray
            scores

        """
        check_variants(variants=variants)
        # log givens for DEBUG == True
        if DEBUG is True:
            print(f'[DEBUG] givens: {givens}')
        # TODO should chooser be settable from the outside ?
        if self.chooser is None:  # add async support
            return DecisionModel._generate_descending_gaussians(count=len(variants))

        try:
            scores = \
                self.chooser.score(
                    variants=variants, givens=givens) + \
                np.array(np.random.rand(len(variants)), dtype='float64') * \
                self.TIEBREAKER_MULTIPLIER
        except Exception as exc:
            warnings.warn(
                'Error when calculating predictions: {}. Returning Gaussian scores'
                .format(exc))
            scores = DecisionModel._generate_descending_gaussians(count=len(variants))
        return scores.astype(np.float64)

    @staticmethod
    def _validate_variants_and_scores(
            variants: list or np.ndarray, scores: list or np.ndarray) -> bool:
        """
        Check if variants and scores are not malformed

        Parameters
        ----------
        variants: np.ndarray
            array of variants
        scores: np.ndarray
            array of scores

        Returns
        -------
        bool
            Flag indicating whether variants and scores are valid

        """

        if variants is None or scores is None:
            raise ValueError('`variants` and `scores` can`t be None')

        if len(variants) != len(scores):
            raise ValueError('Lengths of `variants` and `scores` mismatch!')

        if len(variants) == 0:
            return False

        return True

    @staticmethod
    def top_scoring_variant(
            variants: list or np.ndarray, scores: list or np.ndarray) -> dict:
        """
        Gets best variant considering provided scores

        Parameters
        ----------
        variants: np.ndarray
            collection of variants to be ranked
        scores: np.ndarray
            collection of scores used for ranking

        Returns
        -------
        dict
            Returns best variant

        """

        assert variants is not None and scores is not None
        # null if no variants for get() implementation
        if not DecisionModel._validate_variants_and_scores(
                variants=variants, scores=scores):
            warnings.warn('The variants of length 0 were provided. Returning None')
            return None

        return variants[np.argmax(scores)]

    @staticmethod
    def rank(
            variants: list or np.ndarray, scores: list or np.ndarray) -> list or np.ndarray:
        """
        Return a list of the variants ranked from best to worst.
        DO NOT USE THIS METHOD - WILL LIKELY CHANGE SOON

        Parameters
        ----------
        variants: np.ndarray
            collection of variants to be ranked
        scores: np.ndarray
            collection of scores used for ranking

        Returns
        -------
        np.ndarray
            sorted variants

        """

        assert variants is not None and scores is not None
        # null if no variants for get() implementation
        if not DecisionModel._validate_variants_and_scores(
                variants=variants, scores=scores):
            return None

        variants_w_scores = np.array([variants, scores], dtype=object).T

        sorted_variants_w_scores = \
            variants_w_scores[(variants_w_scores[:, 1] * -1).argsort()]
        return sorted_variants_w_scores[:, 0]

    @staticmethod
    def _generate_descending_gaussians(count: int) -> list or np.ndarray:
        """
        Generates random floats and sorts in a descending fashion

        Parameters
        ----------
        count: int
            number of floats to generate

        Returns
        -------
        np.ndarray
            array of sorted floats

        """

        random_scores = np.random.normal(size=count)
        random_scores[::-1].sort()
        return random_scores

    def given(self, givens: dict or None) -> dc.DecisionContext:  # returns DecisionContext?
        """
        Wrapper for chaining.

        Parameters
        ----------
        givens: dict
            givens to be set

        Returns
        -------

        """

        return dc.DecisionContext(decision_model=self, givens=givens)

    def choose_from(self, variants: list):
        """
        Wrapper for chaining

        Parameters
        ----------
        variants: np.ndarray
            variants to be set

        Returns
        -------
        Decision
            decision object with provided with already scored variants and best selected

        """
        check_variants(variants=variants)
        # TODO how about the givens? Should GivensProvider be used here?
        #  in iOS here givens are set to nil
        return dc.DecisionContext(decision_model=self, givens=None).choose_from(variants=variants)

    def which(self, *variants):
        """
        A short hand version of chooseFrom that returns the chosen result directly,
        automatically tracking the decision in the process. This would be a varargs
        version of chooseFrom that would also take an array as the only argument.
        If the only argument was not an array it would throw an exception.
        Since the Decision object is not returned, rewards would only be tracked
        through DecisionModel.addReward

        Parameters
        ----------
        *variants: list or tuple or np.ndarray
            array of variants passed as positional parameters

        Returns
        -------
        d.Decision
            snapshot of the Decision made with provided variants and available givens

        """
        return self.choose_from(variants=get_variants_for_which(variants=variants)).get()



from ksuid import Ksuid
import numpy as np
from warnings import warn

import improveai.decision_model as dm
from improveai.utils.general_purpose_tools import check_variants


class Decision:

    @property
    def variants(self) -> list or np.ndarray or tuple:
        return self.__variants

    @variants.setter
    def variants(self, value: list or np.ndarray or tuple):
        # TODO test variants setter behaviour
        if self.__variants_set is False:
            check_variants(variants=value)
            self.__variants = value
            self.__variants_set = True

    @property
    def givens(self) -> dict or None:
        return self.__givens

    @givens.setter
    def givens(self, value: dict):
        # TODO test givens setter behaviour
        if self.__givens_set is False:
            assert isinstance(value, dict) or value is None
            self.__givens = value
            self.__givens_set = True

    @property
    def decision_model(self) -> object:
        return self.__decision_model

    @property
    def chosen(self):
        return self.__chosen

    @property
    def best(self):
        return self.__best

    @best.setter
    def best(self, value):
        # TODO test best setter behaviour (previously memoized_variant)
        if self.__best_set is False:
            self.__best = value
            self.__best_set = True

    @property
    def scores(self):
        return self.__scores

    @scores.setter
    def scores(self, value):
        # TODO test scores setter behaviour
        if self.__scores_set is False:
            self.__scores = value
            self.__scores_set = True

    @property
    def ranked_variants(self):
        return self.__ranked_variants

    @property
    def id_(self):
        return self.__id_

    def __init__(self, decision_model: object):

        assert isinstance(decision_model, dm.DecisionModel)

        self.__decision_model = decision_model
        self.__chosen = False

        self.__variants = [None]
        self.__variants_set = False

        self.__givens = None
        self.__givens_set = False

        self.__best = None
        self.__best_set = False

        self.__scores = None
        self.__scores_set = False

        self.__ranked_variants = [None]
        self.__id_ = None

    def _set_message_id(self):
        if self.__id_ is None:
            id_ = str(Ksuid())
            assert isinstance(id_, str)
            Ksuid.from_base62(id_)
            self.__id_ = id_
        else:
            warn('`id_` has already been set for this Decision')

    def _cache_message_id_to_decision_model(self):
        assert self.__id_ is not None
        self.decision_model.id_ = self.id_

    def add_reward(self, reward: float or int):

        if not self.chosen:
            warn('`add_reward()` called before `get()`')

        assert self.decision_model.id_ == self.id_
        assert self.decision_model.model_name is not None and self.decision_model.id_ is not None
        assert reward is not None
        assert isinstance(reward, float) or isinstance(reward, int)
        assert not np.isnan(reward)
        assert not np.isinf(reward)

        return self.decision_model._tracker.add_reward(
            reward=reward, model_name=self.decision_model.model_name, decision_id=self.id_)

    def get(self):

        if self.chosen:
            warn('The best variant has already been chosen')
            self._cache_message_id_to_decision_model()
            return self.best

        # set message_id / decision_id only once
        self._set_message_id()
        # set message_id / deicsion_id to decision model
        self._cache_message_id_to_decision_model()

        self.__scores = self.decision_model._score(variants=self.variants, givens=self.givens)

        if self.variants is not None and len(self.variants) != 0:
            # there should be no difference between effect of those 2 conditions
            # since this  clause is reached only once
            if self.decision_model._tracker:

                track_runners_up = self.decision_model._tracker._should_track_runners_up(len(self.variants))
                if track_runners_up:
                    self.__ranked_variants = \
                        dm.DecisionModel.rank(variants=self.variants, scores=self.__scores)
                    self.__best = self.ranked_variants[0]

                    self.decision_model._tracker.track(
                        variant=self.best, variants=self.ranked_variants,
                        givens=self.givens, model_name=self.decision_model.model_name,
                        variants_ranked_and_track_runners_up=True, message_id=self.id_)
                else:
                    self.__best = \
                        dm.DecisionModel.top_scoring_variant(variants=self.variants, scores=self.__scores)

                    self.decision_model._tracker.track(
                        variant=self.best, variants=self.variants,
                        givens=self.givens, model_name=self.decision_model.model_name,
                        variants_ranked_and_track_runners_up=False, message_id=self.id_)

            elif self.decision_model._tracker is None:
                raise ValueError('`tracker` object can`t be None')
            else:
                self.__best = \
                    dm.DecisionModel.top_scoring_variant(
                        variants=self.variants, scores=self.__scores)

        self.__chosen = True
        return self.best



import numpy as np

import improveai.decision as d
import improveai.decision_model as dm
import improveai.givens_provider as gp
from improveai.utils.general_purpose_tools import check_variants, get_variants_for_which


class DecisionContext:

    @property
    def decision_model(self):
        return self._decision_model

    @decision_model.setter
    def decision_model(self, value):
        # DecisionContext's model must not be None and of desired type
        assert value is not None
        assert isinstance(value, dm.DecisionModel)
        self._decision_model = value

    @property
    def givens(self):
        return self._givens

    @givens.setter
    def givens(self, value):
        # givens can be None or dict
        assert isinstance(value, dict) or value is None
        self._givens = value

    # TODO should givens be a part of constructor call (they are in iOS im implementation)
    def __init__(self, decision_model, givens: dict or None):
        self.decision_model = decision_model
        self.givens = givens

    def choose_from(self, variants):
        """
        Makes a Decision without tracking it

        Parameters
        ----------
        variants: list or tuple or np.ndarray
            array of variants to choose from

        Returns
        -------
        Decision
            snapshot of the Decision made with provided variants and available givens

        """
        check_variants(variants=variants)
        # givens must be provided at this point -> they are needed for decision snapshot
        givens = gp.GivensProvider().givens(for_model=self.decision_model, givens=self.givens)
        # calculate scores (with _score() method) -> givens are provided inside it
        # scores will also be cached to decision object
        scores = self.decision_model._score(variants=variants, givens=givens)
        # calculate the best variant without tracking (calling get())
        best = self.decision_model.top_scoring_variant(variants=variants, scores=scores)
        # cache all info to decision object
        decision = d.Decision(decision_model=self.decision_model)
        decision.variants = variants
        decision.givens = givens
        decision.scores = scores
        decision.best = best

        return decision

    def score(self, variants):
        """
        Scores provided variants with available givens

        Parameters
        ----------
        variants: list or tuple or np.ndarray
            array of variants to choose from

        Returns
        -------
        np.ndarray
            array of float64 scores

        """
        # TODO how givens should be provided here?
        # I assumed that this should be done like in iOS SDK
        givens = gp.GivensProvider().givens(for_model=self.decision_model, givens=self.givens)
        return self.decision_model._score(variants=variants, givens=givens)

    def which(self, *variants):
        """
        A short hand version of chooseFrom that returns the chosen result directly,
        automatically tracking the decision in the process. This would be a varargs
        version of chooseFrom that would also take an array as the only argument.
        If the only argument was not an array it would throw an exception.
        Since the Decision object is not returned, rewards would only be tracked
        through DecisionModel.addReward

        Parameters
        ----------
        *variants: list or tuple or np.ndarray
            array of variants passed as positional parameters

        Returns
        -------
        Decision
            snapshot of the Decision made with provided variants and available givens

        """
        return self.choose_from(variants=get_variants_for_which(variants=variants)).get()



from collections.abc import Iterable
from copy import deepcopy
import json
import numpy as np
import pickle
import re
from time import time
from typing import Dict, List
from traceback import print_exc
from xgboost import Booster, DMatrix
from xgboost.core import XGBoostError

from improveai.feature_encoder import FeatureEncoder
from improveai.choosers.basic_choosers import BasicChooser
# import improveai.cythonized_feature_encoding.cythonized_feature_encoder as cfe
# import improveai.cythonized_feature_encoding.cythonized_feature_encoding_utils as cfeu
from improveai.cythonized_feature_encoding import cfe, cfeu
from improveai.settings import USE_CYTHON_BACKEND
from improveai.utils.general_purpose_tools import constant
from improveai.utils.choosers_feature_encoding_tools import encoded_variants_to_np


FastFeatureEncoder = cfe.FeatureEncoder


class NativeXGBChooser(BasicChooser):

    @property
    def model(self) -> Booster:
        return self._model

    @model.setter
    def model(self, new_val: Booster):
        self._model = new_val

    @property
    def model_metadata(self) -> Dict[str, object]:
        return self._model_metadata

    @model_metadata.setter
    def model_metadata(self, new_val: Dict[str, object]):
        self._model_metadata = new_val

    @property
    def feature_encoder(self) -> FeatureEncoder or FastFeatureEncoder:
        return self._feature_encoder

    @feature_encoder.setter
    def feature_encoder(self, new_val: FeatureEncoder or FastFeatureEncoder):
        self._feature_encoder = new_val

    @property
    def model_metadata_key(self):
        return self._mlmodel_metadata_key

    @model_metadata_key.setter
    def model_metadata_key(self, new_val: str):
        self._mlmodel_metadata_key = new_val

    @property
    def model_seed_key(self) -> str:
        return self._model_seed_key

    @model_seed_key.setter
    def model_seed_key(self, new_val: str):
        self._model_seed_key = new_val

    @property
    def model_seed(self):
        return self._model_seed

    @model_seed.setter
    def model_seed(self, value):
        self._model_seed = value

    @property
    def model_name_key(self):
        return self._model_name_key

    @model_name_key.setter
    def model_name_key(self, value):
        self._model_name_key = value

    @property
    def model_name(self):
        return self._model_name

    @model_name.setter
    def model_name(self, value):
        # TODO if it will be decided that model_name can ne None refactor
        assert value is not None
        # if value is not None:
        assert isinstance(value, str)
        assert re.search(NativeXGBChooser.MODEL_NAME_REGEXP, value) is not None

        self._model_name = value

    @property
    def model_feature_names_key(self) -> str:
        return self._model_feature_names_key

    @model_feature_names_key.setter
    def model_feature_names_key(self, new_val: str):
        self._model_feature_names_key = new_val

    @property
    def model_feature_names(self) -> np.ndarray:
        return self._model_feature_names

    @model_feature_names.setter
    def model_feature_names(self, new_val: np.ndarray):
        self._model_feature_names = new_val

    @property
    def feature_encoder_extras(self):
        return self._feature_encoder_extras

    @feature_encoder_extras.setter
    def feature_encoder_extras(self, value):
        self._feature_encoder_extras = value

    @property
    def current_noise(self):
        return self._current_noise

    @current_noise.setter
    def current_noise(self, value):
        self._current_noise = value

    @property
    def imposed_noise(self):
        return self._imposed_noise

    @imposed_noise.setter
    def imposed_noise(self, value):
        # assert noise is valid
        assert not isinstance(value, bool)
        assert isinstance(value, float) or isinstance(value, int)
        assert 0 <= value <= 1
        self._imposed_noise = value

    @constant
    def SUPPORTED_OBJECTIVES() -> list:
        return ['reg', 'binary', 'multi']

    def __init__(
            self, mlmodel_metadata_key: str = 'json',
            model_feature_names_key: str = 'feature_names',
            model_seed_key: str = 'model_seed',
            model_name_key: str = 'model_name'):

        self.model = None
        self.model_metadata_key = mlmodel_metadata_key
        self.model_metadata = None

        self.feature_encoder = None
        self.model_feature_names_key = model_feature_names_key
        self.model_feature_names = np.empty(shape=(1,))

        self.model_seed_key = model_seed_key
        self.model_seed = None

        self.model_name_key = model_name_key
        self._model_name = None

        self.current_noise = None
        self._imposed_noise = None

    def load_model(self, input_model_src: str, verbose: bool = False):
        """
        Loads desired model from input path.

        Parameters
        ----------
        input_model_src: str
            path to desired model
        verbose: bool
            should I print msgs
        Returns
        -------
        None
            None
        """

        try:
            if verbose:
                print('Attempting to load: {} model'.format(
                    input_model_src if len(input_model_src) < 100 else
                    str(input_model_src[:10]) + ' ... ' + str(
                        input_model_src[-10:])))

            raw_model_src = self.get_model_src(model_src=input_model_src)

            model_src = \
                raw_model_src if isinstance(raw_model_src, str) \
                else bytearray(raw_model_src)

            self.model = Booster()
            self.model.load_model(model_src)
            if verbose:
                print('Model: {} successfully loaded'.format(
                    input_model_src if len(input_model_src) < 100 else
                    str(input_model_src[:10]) + ' ... ' + str(
                        input_model_src[-10:])))
        except XGBoostError as xgbe:
            if verbose:
                print('Attempting to read via pickle interface')
            with open(input_model_src, 'rb') as xgbl:
                self.model = pickle.load(xgbl)
            print_exc()
        except Exception as exc:
            if verbose:
                print(
                    'When attempting to load the model: {} the following error '
                    'occured: {}'.format(input_model_src, exc))
            print_exc()

        model_metadata = self._get_model_metadata()
        self.model_seed = self._get_model_seed(model_metadata=model_metadata)
        self.model_name = self._get_model_name(model_metadata=model_metadata)
        self.model_feature_names = \
            self._get_model_feature_names(model_metadata=model_metadata)

        if USE_CYTHON_BACKEND:
            self.feature_encoder = FastFeatureEncoder(model_seed=self.model_seed)
        else:
            self.feature_encoder = FeatureEncoder(model_seed=self.model_seed)

    def _get_model_metadata(self) -> dict:
        """
        Model metadata (hash table, etc. getter)

        Returns
        -------
        Dict[str, object]
            dict with model metadata
        """

        assert 'user_defined_metadata' in self.model.attributes().keys()
        user_defined_metadata_str = self.model.attr('user_defined_metadata')
        user_defined_metadata = json.loads(user_defined_metadata_str)
        assert self.model_metadata_key in user_defined_metadata.keys()

        return user_defined_metadata[self.model_metadata_key]

    def score(
            self, variants: List[Dict[str, object]],
            givens: Dict[str, object], imputer_value: float = np.nan,
            **kwargs) -> np.ndarray:

        """
        Scores all provided variants

        Parameters
        ----------
        variants: list
            list of variants to scores
        givens: dict
            context dict needed for encoding
                mlmodel_score_res_key
        imputer_value: float
            value with which missing valuse will be imputed
        kwargs

        Returns
        -------
        np.ndarray
            2D numpy array which contains (variant, score) pair in each row
        """

        encoded_variants = \
            self._encode_variants_single_givens(variants=variants, givens=givens)

        encoded_variants_to_np_method = \
            cfeu.encoded_variants_to_np if USE_CYTHON_BACKEND \
            else encoded_variants_to_np

        missings_filled_v = \
            encoded_variants_to_np_method(
                encoded_variants=encoded_variants,
                feature_names=self.model_feature_names)

        if USE_CYTHON_BACKEND:
            missings_filled_v = np.asarray(missings_filled_v)

        scores = \
            self.model.predict(
                DMatrix(
                    missings_filled_v, feature_names=self.model_feature_names))\
            .astype('float64')

        # scores += \
        #     np.array(
        #         np.random.rand(len(encoded_variants)), dtype='float64') * \
        #     self.TIEBREAKER_MULTIPLIER

        return scores

    def _encode_variants_single_givens(
            self, variants: Iterable, givens: dict or None) -> Iterable:
        """
        Implemented as a XGBChooser helper method
        Cythonized loop over provided variants and a single givens dict.
        Returns array of encoded dicts.
        Parameters
        ----------
        variants: Iterable
            collection of input variants to be encoded
        givens: dict or None
            context to be encoded with variants

        Returns
        -------
        np.ndarray
            array of encoded dicts
        """

        if not (isinstance(givens, dict) or givens is None or givens is {}):
            raise TypeError(
                'Unsupported givens` type: {}'.format(type(givens)))
            # process with single context

        if self.imposed_noise is None:
            noise = np.random.rand()
        else:
            noise = self.imposed_noise
        # self.current_noise = noise = np.random.rand()

        if USE_CYTHON_BACKEND:
            if isinstance(variants, list):
                used_variants = variants
            elif isinstance(variants, tuple) or isinstance(variants, np.ndarray):
                used_variants = list(variants)
            else:
                raise TypeError(
                    'Variants are of a wrong type: {}'.format(type(variants)))

            return cfeu.encode_variants_single_givens(
                variants=used_variants, givens=givens, noise=noise,
                variants_encoder=self.feature_encoder.encode_variant,
                givens_encoder=self.feature_encoder.encode_givens)
        else:
            encoded_variants = np.empty(len(variants), dtype=object)

            encoded_givens = self.feature_encoder.encode_givens(givens=givens, noise=noise)

            encoded_variants[:] = [
                self.feature_encoder.encode_variant(
                    variant=variant, noise=noise, into=deepcopy(encoded_givens))
                for variant in variants]

            return encoded_variants


if __name__ == '__main__':

    mlmc = NativeXGBChooser()

    # test_model_pth = '../artifacts/models/12_11_2020_verses_conv.xgb'
    # test_model_pth = "https://improve-v5-resources-prod-models-117097735164.s3-us-west-2.amazonaws.com/models/mindful/latest/improve-stories-2.0.xgb.gz"
    test_model_pth = "/Users/os/Downloads/model.gz"
    test_model_pth = \
        '/home/kw/Projects/upwork/python-sdk/improveai/tests/artifacts/models/dummy_v6.xgb'
    mlmc.load_model(input_model_src=test_model_pth)

    # with open('../artifacts/test_artifacts/model.json', 'r') as mj:
    #     json_str = mj.readline()
    #     model_metadata = json.loads(json_str)

    with open('../tests/artifacts/test_artifacts/context.json', 'r') as mj:
        json_str = mj.readline()
        context = json.loads(json_str)

    with open('../tests/artifacts/test_artifacts/meditations.json', 'r') as vj:
        json_str = vj.readlines()
        variants = json.loads(''.join(json_str))

    # features_count = \
    #     len(model_metadata["table"][1])
    # feature_names = list(
    #     map(lambda i: 'f{}'.format(i), range(0, features_count)))
    #
    # sample_variants = [{"arrays": [1 for el in range(0, features_count + 10)]},
    #                    {"arrays": [el + 2 for el in range(0, features_count + 10)]},
    #                    {"arrays": [el for el in range(0, features_count + 10)]}]
    # 0.3775409052734039

    # single_score = mlmc.score(
    #     variant=variants[0], givens=context, sigmoid_correction=True)
    # print('single score')
    # print(single_score)

    st = time()
    batch_size = 100
    for _ in range(batch_size):
        score_all = mlmc.score(
            variants=variants, givens=context, sigmoid_correction=True,
            return_plain_results=True)
        score_all[-1] = 100
        score_all[::-1].sort()
        best = score_all[0]
    et = time()
    print(score_all[:10])
    print((et - st) / batch_size)
    input('score_all')

    score_all = mlmc.score(
        variants=variants, givens=context)

    sorted = mlmc.sort(variants_w_scores=score_all)
    print('sorted')
    print(sorted)

    choose = mlmc.choose(variants_w_scores=score_all)
    print('choose')
    print(choose)



import coremltools as ct
import json
import numpy as np
from time import time
from typing import Dict, List


from improveai.feature_encoder import FeatureEncoder
# from improveai.cythonized_feature_encoding import cfe
import improveai.cythonized_feature_encoding.cythonized_feature_encoder as cfe
from improveai.choosers.basic_choosers import BasicChooser
from improveai.settings import USE_CYTHON_BACKEND
from improveai.utils.general_purpose_tools import constant
from improveai.utils.choosers_feature_encoding_tools import \
    encoded_variant_to_np

FastFeatureEncoder = cfe.FeatureEncoder


class MLModelChooser(BasicChooser):

    @property
    def model(self) -> ct.models.MLModel:
        return self._model

    @model.setter
    def model(self, new_val: ct.models.MLModel):
        self._model = new_val

    @property
    def model_metadata(self) -> Dict[str, object]:
        return self._model_metadata

    @model_metadata.setter
    def model_metadata(self, new_val: Dict[str, object]):
        self._model_metadata = new_val

    @property
    def feature_encoder(self) -> FeatureEncoder or FastFeatureEncoder:
        return self._feature_encoder

    @feature_encoder.setter
    def feature_encoder(self, new_val: FeatureEncoder or FastFeatureEncoder):
        self._feature_encoder = new_val

    @property
    def model_metadata_key(self) -> str:
        return self._model_metadata_key

    @model_metadata_key.setter
    def model_metadata_key(self, new_val: str):
        self._model_metadata_key = new_val

    @property
    def model_seed_key(self) -> str:
        return self._model_seed_key

    @model_seed_key.setter
    def model_seed_key(self, new_val: str):
        self._model_seed_key = new_val

    @property
    def model_seed(self):
        return self._model_seed

    @model_seed.setter
    def model_seed(self, value):
        self._model_seed = value

    @property
    def model_name_key(self):
        return self._model_name_key

    @model_name_key.setter
    def model_name_key(self, value):
        self._model_name_key = value

    @property
    def model_name(self):
        return self._model_name

    @model_name.setter
    def model_name(self, value):
        self._model_name = value

    @property
    def model_feature_names_key(self):
        return self._model_feature_names_key

    @model_feature_names_key.setter
    def model_feature_names_key(self, new_val):
        self._model_feature_names_key = new_val

    @property
    def model_feature_names(self) -> np.ndarray:
        return self._model_feature_names

    @model_feature_names.setter
    def model_feature_names(self, new_val):
        self._model_feature_names = new_val

    @constant
    def TIEBREAKER_MULTIPLIER() -> float:
        return 2e-23

    def __init__(
            self, mlmodel_metadata_key: str = 'json',
            model_feature_names_key: str = 'feature_names',
            model_seed_key: str = 'model_seed',
            model_name_key: str = 'model_name'):
        # initialize
        self.model = None
        self.model_metadata_key = mlmodel_metadata_key
        self.model_metadata = None

        self.feature_encoder = None

        self.model_seed_key = model_seed_key
        self.model_seed = None

        self.model_name_key = model_name_key
        self.model_name = None

        self.model_objective = None

        self.model_feature_names_key = model_feature_names_key
        self.model_feature_names = np.empty(shape=(1,))

    def _load_buffered_model(self, model_bytes: bytes):
        """
        # TODO make this in memory 
        Loads mlmodel using temp cache file  
        
        Parameters
        ----------
        model_bytes: bytes
            model bytes

        Returns
        -------

        """

        spec = ct.proto.Model_pb2.Model()
        spec.ParseFromString(model_bytes)
        return ct.models.MLModel(spec)

    def load_model(self, input_model_src: str, verbose: bool = False):
        """
        Loads desired model from input path.

        Parameters
        ----------
        input_model_src: str
            path to desired model
        verbose: bool
            should I print msgs

        Returns
        -------
        None
            None

        """

        failed_to_load = False
        try:
            if verbose:
                print('Attempting to load: {} model'.format(
                    input_model_src if len(input_model_src) < 100 else
                    str(input_model_src[:10]) + ' ... ' + str(
                        input_model_src[-10:])))

            raw_model_src = self.get_model_src(model_src=input_model_src)

            # model_src = self._load_buffered_model(model_bytes=raw_model_src)

            self.model = \
                ct.models.MLModel(raw_model_src) \
                if not isinstance(raw_model_src, bytes) \
                else self._load_buffered_model(model_bytes=raw_model_src)

            if verbose and not failed_to_load:
                print('Model: {} successfully loaded'.format(
                    input_model_src if len(input_model_src) < 100 else
                    str(input_model_src[:10]) + ' ... ' + str(input_model_src[-10:])))
        except Exception as exc:
            if verbose:
                print(
                    'When attempting to load the mode: {} the following error '
                    'occured: {}'.format(input_model_src, exc))

        if failed_to_load:
            raise RuntimeError('Model failed to load')

        model_metadata = self._get_model_metadata()

        self.model_seed = self._get_model_seed(model_metadata=model_metadata)
        self.model_name = self._get_model_name(model_metadata=model_metadata)
        self.model_feature_names = \
            self._get_model_feature_names(model_metadata=model_metadata)

        if USE_CYTHON_BACKEND:
            self.feature_encoder = FastFeatureEncoder(model_seed=self.model_seed)
        else:
            self.feature_encoder = FeatureEncoder(model_seed=self.model_seed)

    def _get_model_metadata(self) -> dict:
        """
        Gets model metadata either from dict or from model attributes

        Returns
        -------
        dict
            metadata dict

        """

        assert hasattr(self.model, 'user_defined_metadata')
        assert self.model_metadata_key in self.model.user_defined_metadata.keys()
        return json.loads(
            self.model.user_defined_metadata[self.model_metadata_key])

    def _score(
            self, variant: Dict[str, object], noise: float,
            givens: Dict[str, object] = None,
            encoded_givens: Dict[str, object] = None, **kwargs) -> list:
        """
        Performs scoring of a single variant using provided context and loaded
        model

        Parameters
        ----------
        variant: dict
            scored case / row as a dict
        givens: dict
            dict with lookup table and seed
        encoded_givens: Dict[str, float]
            already encoded context dict
        kwargs

        Returns
        -------
        np.ndarray
            score of provided variant

        """

        if not encoded_givens:
            encoded_givens = \
                self.feature_encoder.encode_givens(givens=givens, noise=noise)

        encoded_variant_and_context = \
            self.feature_encoder.encode_variant(
                variant=variant, noise=noise, into=encoded_givens)

        missings_filled_v = \
            encoded_variant_to_np(
                encoded_variant=encoded_variant_and_context,
                feature_names=self.model_feature_names)

        assert len(self.model_feature_names) == len(missings_filled_v)

        score_dict = \
            self.model.predict(
                dict(zip(self.model_feature_names, missings_filled_v)))

        if 'target' not in score_dict:
            raise ValueError(
                'Prediction dict has no `target` key: {}'.format(score_dict))

        score = score_dict.get('target', None)

        return score

    def score(
            self, variants: List[Dict[str, object]],
            givens: Dict[str, object], **kwargs) -> np.ndarray:
        """
        Scores all provided variants

        Parameters
        ----------
        variants: list
            list of variants to scores
        givens: dict
            dict with lookup table and seed
        kwargs
            kwargs

        Returns
        -------
        np.ndarray
            2D numpy array which contains (variant, score) pair in each row

        """

        # encoded_context = \
        #     self.feature_encoder.encode_features({'context': context})

        noise = np.random.rand()

        encoded_context = \
            self.feature_encoder.encode_givens(givens=givens, noise=noise)

        scores = \
            np.array([self._score(
                variant=variant, noise=noise, givens=givens,
                encoded_givens=encoded_context) for variant in variants])

        return scores


if __name__ == '__main__':

    mlmc = MLModelChooser()

    # test_model_pth = '../tests/artifacts/test_artifacts/improve-messages-2.0-3.mlmodel'
    test_model_pth = '../tests/artifacts/models/v6_conv_model.mlmodel'
    # test_model_pth = "https://improve-v5-resources-prod-models-117097735164.s3-us-west-2.amazonaws.com/models/mindful/latest/improve-messages-2.0.mlmodel"
    # test_model_pth = "/Users/os/Downloads/improve-messages-2.0.mlmodel.gz"
    mlmc.load_model(input_model_src=test_model_pth)

    with open('../tests/artifacts/test_artifacts/context.json', 'r') as mj:
        json_str = mj.readline()
        context = json.loads(json_str)

    with open('../tests/artifacts/test_artifacts/meditations.json', 'r') as vj:
        json_str = vj.readlines()
        variants = json.loads(''.join(json_str))

    noise = np.random.rand()
    res = mlmc._score(variant=variants[0], givens=context, noise=noise)
    print('res')
    print(res)
    # 0.0012913734900291936
    # input('check')

    batch_size = 10
    print(len(variants))
    st = time()

    for _ in range(batch_size):
        # res = mlmc.score(variant=variants[0], context=context)
        res_all = \
            mlmc.score(variants=variants, givens=context)
    et = time()
    print((et - st) / batch_size)
    input('sanity check')
    print('res_all')

    srtd_variants_w_scores = mlmc.sort(variants_w_scores=res_all)
    print('srtd_variants_w_scores')
    for row in srtd_variants_w_scores.tolist():
        print('{} -> sigmoid({}) = {}'.format(row, row[1], round(1 / (1 + np.exp(0.5 - row[1])), 4)))

    best_choice = mlmc.choose(variants_w_scores=srtd_variants_w_scores)
    print('best_choice')
    print(best_choice)


# {28: 1.0, 33: 1.0, 17: 1.0, 39: 1.0, 42: 1.0, 11: 1.0, 37: 1.0, 9: 1.0, 31: 1.0, 38: 1.0, 41: 1.0, 21: 1.0, 3: 1.0, 43: 1.0, 22: 1.0, 2: 1.0, 8: 1.0, 14: 1.0, 40: 1.0, 25: 1.0, 20: 1.0, 30: 1.0, 16: 1.0, 4: 1.0, 19: 1.0, 6: 1.0, 23: 1.0, 26: 1.0, 24: 1.0, 27: 1.0, 18: 1.0, 29: 1.0, 10: 1.0, 1: 1.0, 32: 1.0, 13: 1.0, 36: 1.0, 34: 1.0, 7: 1.0, 0: 1.0, 12: 1.0, 35: 1.0, 15: 1.0, 5: 1.0}
# {28: 1.0, 33: 1.0, 17: 1.0, 39: 1.0, 42: 1.0, 11: 1.0, 37: 1.0, 9: 1.0, 31: 1.0, 38: 1.0, 41: 1.0, 21: 1.0, 3: 1.0, 43: 1.0, 22: 1.0, 2: 1.0, 8: 1.0, 14: 1.0, 40: 1.0, 25: 1.0, 20: 1.0, 30: 1.0, 16: 1.0, 4: 1.0, 19: 1.0, 6: 1.0, 23: 1.0, 26: 1.0, 24: 1.0, 27: 1.0, 18: 1.0, 29: 1.0, 10: 1.0, 1: 1.0, 32: 1.0, 13: 1.0, 36: 1.0, 34: 1.0, 7: 1.0, 0: 1.0, 12: 1.0, 35: 1.0, 15: 1.0, 5: 1.0}



from codecs import decode
from copy import deepcopy
import numpy as np


def constant(f):
    """
    Constant decorator raising error on attempt to set decorated value

    Parameters
    ----------
    f: callable
        decorated object

    Returns
    -------
    property
        constant property

    """
    def fset(self, value):
        raise AttributeError

    def fget(self):
        return f()

    return property(fget, fset)


def append_prfx_to_dict_keys(input_dict: dict, prfx: str) -> dict:
    """
    Appends prefix to the input dict keys

    Parameters
    ----------
    input_dict: dict
        dict which keys will be appended
    prfx: str
        prf to append to input dict keys

    Returns
    -------
    dict
        dict with keys appended

    """
    return dict(zip(
        ['{}{}'.format(prfx, key) for key in input_dict.keys()],
        list(input_dict.values())))


def impute_missing_dict_keys(
        all_des_keys: list, imputed_dict, imputer_value: float = np.nan):
    """
    Imputes provided value into missing keys of input dict

    Parameters
    ----------
    all_des_keys: list
        list of all necessary keys in returned dict
    imputed_dict: dict
        input dict
    imputer_value: object
        value to be imputed for missing values

    Returns
    -------
    dict
        dict with all missing keys imputed

    """

    prcsd_encoded_features = deepcopy(imputed_dict)

    for feat_n in all_des_keys:
        if feat_n not in imputed_dict.keys():
            prcsd_encoded_features[feat_n] = imputer_value

    return prcsd_encoded_features


def read_jsonstring_from_file(
        path_to_file: str, mode: str = 'r', method: str = 'readlines',
        decode_escape_char: bool = False) -> str:
    """
    Safely reads desired file with json wrapper

    Parameters
    ----------
    path_to_file: str
        pth to loaded file
    mode: str
        read mode
    decode_escape_char: bool
        should double escape be repalced by single ?

    Returns
    -------
    str
        read json string

    """

    with open(path_to_file, mode) as rf:
        if method == 'readlines':
            contents = rf.readlines()
        elif method == 'read':
            contents = rf.read()
        else:
            raise ValueError('Wrong reading method value: {}'.format(method))

    if isinstance(contents, list):
        contents = ''.join(contents)

    if decode_escape_char:
        contents = decode(contents, 'unicode_escape')

    assert isinstance(contents, str)
    return contents


def sigmoid(x: float, logit_const: float) -> float:
    """
    Calculate sigmoid value

    Parameters
    ----------
    x: float
        arg passed to sigmoid function
    logit_const: float
        constant added to sigmoid argument

    Returns
    -------
    float
        value of sigmoid function

    """
    exp_arg = logit_const - x
    return 1 / (1 + np.exp(exp_arg))


def check_variants(variants: list or tuple or np.ndarray):
    """
    Check if variants are of desired type and if they are not an empty collection

    Parameters
    ----------
    variants: list or tuple or np.ndarray
        checked variants

    """
    assert variants is not None
    assert (isinstance(variants, list) or isinstance(variants, tuple) or isinstance(variants, np.ndarray))
    # raise if variants are an empty list
    if len(variants) == 0:
        raise ValueError('`variants` must not be an empty collection')


def get_variants_for_which(variants: list or tuple or np.ndarray):
    if len(variants) == 1:
        # if the first element is a list -> it is ok to use such list as variants
        # if the first element is not a list / tuple / np.ndarray raise
        assert isinstance(variants[0], list) or isinstance(variants[0], tuple) or isinstance(variants[0], np.ndarray)
        return variants[0]

    return variants




from collections.abc import Iterable
from copy import deepcopy
from datetime import datetime
import json
import numpy as np
import requests as rq
from typing import Dict
from warnings import warn

# todo change when ready
from ksuid import Ksuid

from improveai.utils.general_purpose_tools import constant


class DecisionTracker:

    @constant
    def MODEL_KEY() -> str:
        return "model"

    @constant
    def HISTORY_ID_KEY() -> str:
        return "history_id"

    @constant
    def TIMESTAMP_KEY() -> str:
        return "timestamp"

    @constant
    def MESSAGE_ID_KEY() -> str:
        return "message_id"

    @constant
    def TYPE_KEY() -> str:
        return "type"

    @constant
    def VARIANT_KEY() -> str:
        return "variant"

    @constant
    def GIVENS_KEY() -> str:
        return "givens"

    @constant
    def VARIANTS_COUNT_KEY() -> str:
        return "count"

    @constant
    def VARIANTS_KEY() -> str:
        return "variants"

    @constant
    def SAMPLE_VARIANT_KEY() -> str:
        return "sample_variant"

    @constant
    def REWARD_TYPE() -> str:
        return 'reward'

    @constant
    def REWARD_KEY():
        return 'reward'

    @constant
    def EVENT_KEY() -> str:
        return "event"

    @constant
    def PROPERTIES_KEY() -> str:
        return "properties"

    @constant
    def DECISION_TYPE() -> str:
        return "decision"

    @constant
    def DECISION_ID_KEY() -> str:
        return "decision_id"

    @constant
    def EVENT_TYPE() -> str:
        return "event"

    @constant
    def API_KEY_HEADER() -> str:
        return "x-api-key"

    @constant
    def HISTORY_ID_DEFAULTS_KEY() -> str:
        return "ai.improve.history_id"

    @constant
    def PAYLOAD_FOR_ERROR_KEY():
        return 'ERROR_WITH_PAYLOAD'

    @constant
    def ERROR_CODE_KEY():
        return 'PY_ERROR_MESSAGE'

    @constant
    def REQUEST_ERROR_CODE_KEY():
        return 'REQUEST_ERROR_CODE'

    @constant
    def RUNNERS_UP_KEY() -> str:
        return 'runners_up'

    @constant
    def VARIANTS_COUNT_KEY() -> str:
        return 'count'

    @constant
    def SAMPLE_KEY() -> str:
        return 'sample'

    @property
    def track_url(self) -> str:
        return self._track_url

    @track_url.setter
    def track_url(self, new_val: str):
        self._track_url = new_val

    @property
    def api_key(self) -> str:
        return self._api_key

    @api_key.setter
    def api_key(self, new_val: str):
        self._api_key = new_val

    @property
    def max_runners_up(self):
        return self._max_runners_up

    @max_runners_up.setter
    def max_runners_up(self, new_val):
        self._max_runners_up = new_val

    def __init__(
            self, track_url: str, max_runners_up: int = 50, track_api_key: str = None):

        self.track_url = track_url
        self.api_key = track_api_key

        # TODO determined whether it should be set once per tracker`s life or
        #  with each track request (?)
        self.max_runners_up = max_runners_up

    def _should_track_runners_up(self, variants_count: int):

        if variants_count == 1 or self.max_runners_up == 0:
            return False
        elif variants_count == 2:
            return True
        else:
            return np.random.rand() < 1 / min(
                variants_count - 1, self.max_runners_up)

    def _top_runners_up(self, ranked_variants: list) -> Iterable or None:

        # TODO ask if max_runners_up should indicate index or a count -
        #  I would assume count

        # TODO should this method return [] instead of None ?

        # len(ranked_variants) - 1 -> this will not include last element of
        # collection

        top_runners_up = ranked_variants[
             1:min(len(ranked_variants), self.max_runners_up + 1)]\
            if ranked_variants is not None else None

        # if there is a positive max_runners_up and more than 1 variant there must be at least 1 runner up
        if len(ranked_variants) > 1 and self.max_runners_up > 0:
            assert top_runners_up is not None and len(top_runners_up) > 0

        # if max_runners_up == 0 there are no runners up
        if self.max_runners_up == 0:
            assert not top_runners_up  # None or []

        # If `top_runners_up` == [] return None
        if top_runners_up is None or len(top_runners_up) == 0:
            return None

        returned_top_runners_up = top_runners_up
        if isinstance(top_runners_up, np.ndarray):
            returned_top_runners_up = top_runners_up.tolist()

        return returned_top_runners_up

    def _is_sample_available(self, variants: list or None, runners_up: list):

        variants_count = len(variants)
        runners_up_count = len(runners_up) if runners_up else 0

        if variants_count - runners_up_count - 1 > 0:
            return True

        return False

    def track(
            self, variant: object, variants: list or np.ndarray, givens: dict,
            model_name: str, variants_ranked_and_track_runners_up: bool,
            timestamp: object = None, message_id: str = None):
        """
        Track that variant is causal in the system

        Parameters
        ----------
        variant: object
            chosen variant
        variants: list or np.ndarray
            collection of tracked variants
        givens: dict:
            givens for variants
        model_name: str
            name of model which made the decision (?) / to which observations
            will be assigned
        variants_ranked_and_track_runners_up: bool
            are the variants ranked and runners up should be tracked
        timestamp: object
            when was decision tracked

        Returns
        -------

        """

        if not self.track_url:
            return None

        if isinstance(variants, np.ndarray):
            variants = variants.tolist()

        body = {
            self.TYPE_KEY: self.DECISION_TYPE,
            self.MODEL_KEY: model_name,
            self.VARIANT_KEY: variant,
            self.VARIANTS_COUNT_KEY:
                len(variants) if variants is not None else 1,
            }

        if len(variants) == 2:
            # for 2 variants variants_ranked_and_track_runners_up should be True
            assert variants_ranked_and_track_runners_up
            if variant != variants[0]:
                variants = list(reversed(variants))

        # for variants_ranked_and_track_runners_up == false (max_runners_up == 0)
        # we skip this clause and runners_up == None
        runners_up = None
        if variants is not None and variants != [None] \
                and variants_ranked_and_track_runners_up:
            assert variant == variants[0]

            runners_up = self._top_runners_up(ranked_variants=variants)

            if runners_up is not None:
                body[self.RUNNERS_UP_KEY] = \
                    self._top_runners_up(ranked_variants=variants)

        # If runners_up == None and len(variants) == 2 -> sample should be extracted
        if self._is_sample_available(variants=variants, runners_up=runners_up):
            sample = \
                self.get_sample(
                    variant=variant, variants=variants,
                    track_runners_up=variants_ranked_and_track_runners_up)
            body[self.SAMPLE_KEY] = sample

        if givens is not None:
            body[self.GIVENS_KEY] = givens

        # TODO return decision ID

        # TODO determine what sort of completion_block should be used
        #  For now completion_block() set to 0
        return self.post_improve_request(
            body_values=body,
            block=lambda result, error: (
                warn("Improve.track error: {}".format(error))
                if error else 0, 0), timestamp=timestamp, message_id=message_id)

    def add_reward(self, reward: float or int, model_name: str, decision_id: str):
        # void addReward(double reward)
        # Add rewards for the most recent Decision for this model name, even if
        # that Decision occurred in a previous session. Sets model on the reward
        # record to be equal to the current modelName.
        # NaN, positive infinity, and negative infinity are not allowed and should throw exceptions

        assert model_name is not None and decision_id is not None
        assert isinstance(reward, float) or isinstance(reward, int)
        assert reward is not None
        assert not np.isnan(reward)
        assert not np.isinf(reward)

        body = {
            self.TYPE_KEY: self.REWARD_TYPE,
            self.MODEL_KEY: model_name,
            self.REWARD_KEY: reward,
            self.DECISION_ID_KEY: decision_id}

        return self.post_improve_request(
            body_values=body,
            block=lambda result, error: (
                warn("Improve.track error: {}".format(error)) if error else 0, 0))

    def get_sample(self, variant: object, variants: list, track_runners_up: bool):
        """
        Gets sample from ranked_variants. Takes runenrs up into account

        Parameters
        ----------
        variant: object
            selected variant (first of ranked variants)
        variants: list or np.ndarray
            list of ranked variants
        track_runners_up: bool
            should runners up be tracked ?

        Returns
        -------
        object
            sample

        """

        assert isinstance(track_runners_up, bool)

        if not (isinstance(variants, list) or isinstance(variants, tuple) or isinstance(variants, np.ndarray)):
            raise TypeError(
                'Provided variants are of a wrong type: {}. Only list type is '
                'allowed'.format(type(variants)))

        assert len(variants) > 1
        if len(variants) == 2:
            assert self.max_runners_up == 0

        # TODO If there are no runners up, then sample is a random sample
        #  from variants with just best excluded.
        if not track_runners_up:
            variant_idx = variants.index(variant)
            while True:
                sample_idx = np.random.randint(0, len(variants))
                if variant_idx != sample_idx:
                    break

            return variants[sample_idx]

        assert variant == variants[0]
        # TODO If there are no remaining variants after best and runners up,
        #  then there is no sample.
        last_runner_up_idx = min(len(variants), self.max_runners_up + 1)
        assert last_runner_up_idx < len(variants)

        # TODO If there are runners up, then sample is a random sample from
        #  variants with best and runners up excluded.
        sample = \
            variants[np.random.randint(last_runner_up_idx, len(variants))]
        return sample

    def _is_valid_message_id(self, message_id: str or None) -> bool:
        """
        Check if message_id is  valid

        Parameters
        ----------
        message_id: str
            checked message_id

        Returns
        -------
        bool
            True if message_id otherwise False

        """
        try:
            assert isinstance(message_id, str) or message_id is None
            Ksuid.from_base62(message_id)
            return True
        except:
            return False

    def post_improve_request(
            self, body_values: Dict[str, object], block: callable,
            message_id: str = None, timestamp: object = None):
        """
        Posts request to tracker endpoint

        Parameters
        ----------
        body_values: dict
            dict to be posted
        block: callable
            callable to execute on error
        timestamp: object
            timestamp of request

        Returns
        -------

        """

        headers = {'Content-Type': 'application/json'}

        # TODO unittest this
        if self.api_key:
            headers[self.API_KEY_HEADER] = self.api_key

        assert self._is_valid_message_id(message_id=message_id)

        body = {
            # was checked -> this convention seems to be accurate
            self.TIMESTAMP_KEY:
                timestamp if timestamp else str(np.datetime_as_string(
                    np.datetime64(datetime.now()), unit='ms', timezone='UTC')),
            # TODO check if this is the desired uuid
            self.MESSAGE_ID_KEY: message_id if message_id is not None else str(Ksuid())
        }

        body.update(body_values)

        # serialization is a must have for this requests
        try:
            payload_json = json.dumps(body)

        except Exception as exc:
            warn("Data serialization error: {}\nbody: {}".format(exc, body))
            return None

        error = None
        resp = None

        try:
            resp = \
                rq.post(url=self.track_url, data=payload_json, headers=headers)

        except Exception as exc:
            error = exc

        if not block:
            return None

        if not error and isinstance(resp, rq.models.Response):

            if resp.status_code >= 400:
                user_info = dict(deepcopy(resp.headers))
                user_info[self.REQUEST_ERROR_CODE_KEY] = str(resp.status_code)
                content = json.dumps(body)

                if content:
                    user_info[self.PAYLOAD_FOR_ERROR_KEY] = content \
                        if len(content) <= 1000 else content[:100]

                error = str(error) if not isinstance(error, str) else error
                error += \
                    ' | when attempting to post to ai.improve got error with ' \
                    'code {} and user info: {}' \
                    .format(str(resp.status_code), json.dumps(user_info))

        json_object = None
        if not error:
            # TODO body might contain objects which are not JSON-encodable
            json_object = json.dumps(body)

        if error:
            print('error')
            print(error)
            block(None, error)
        elif json_object:
            block(json_object, None)
        else:
            raise NotImplementedError(
                'Both error and payload objects are None / empty - this should '
                'not happen (?)')

        return resp


if __name__ == '__main__':
    import time
    from tqdm import tqdm

    def post_requests_batch():

        track_url = 'https://trn1jywns6.execute-api.us-east-2.amazonaws.com/track'

        dt = DecisionTracker(track_url=track_url)

        # resp = decision_tracker.track(
        #     variant=variant,
        #     variants=np.array(variants),
        #     givens=givens, model_name=self.dummy_model_name,
        #     variants_ranked_and_track_runners_up=True,
        #     message_id=self.dummy_message_id,
        #     history_id=self.dummy_history_id,
        #     timestamp=self.dummy_timestamp)

        variants = [el for el in range(200)]
        # variants[0] = ''.join(['x' for _ in range(int(10110000/10))])

        # with open('dummy.json', 'w') as dj:
        #     q = json.dumps(variants[0])
        #     dj.write(q)

        # np.random.shuffle(variants)

        import base64

        def base64len(s):
            encoded_str = base64.b64decode(s)
            return len(encoded_str)
        large_variant = ''.join(['x' for _ in range(1011100)])

        # to make rewarding reproducible
        np.random.seed(0)

        persistent_decision_id = None

        for d_idx in tqdm(range(100)):
            # time.sleep(np.random.randint(6,  18))

            givens = {}

            if np.random.rand() > 0.8:
                givens = {
                    'g1': 0,
                    'g2': 1}

            decision_id = str(Ksuid())

            # if not persistent_decision_id:
            #     persistent_decision_id = decision_id

            if d_idx % 15 == 0:
                persistent_decision_id = decision_id

            resp = dt.track(
                variant=variants[0],
                variants=variants[:1],
                givens=givens, model_name='appconfig',
                variants_ranked_and_track_runners_up=False,
                timestamp=str(np.datetime_as_string(
                            np.datetime64(datetime.now()), unit='ms', timezone='UTC')),
                message_id=decision_id)

            # if d_idx < 100:
            # time.sleep(np.random.rand() * 0.2)
            dt.add_reward(reward=100.0, model_name='appconfig', decision_id=decision_id)
            # time.sleep(np.random.rand() * 0.2)
            dt.add_reward(reward=75.0, model_name='appconfig', decision_id=decision_id)
            # time.sleep(np.random.rand() * 0.2)
            dt.add_reward(reward=50.0, model_name='appconfig', decision_id=decision_id)
            # resp = dt.add_reward(reward=1.0, model_name='appconfig', decision_id=decision_id)
            if np.random.rand() < 0.5:
                dt.add_reward(reward=10.0, model_name='appconfig', decision_id=persistent_decision_id)

            print(resp.status_code)
            # time.sleep(np.random.rand() * 0.1)

    from multiprocessing import Process

    p1 = Process(target=post_requests_batch)
    p2 = Process(target=post_requests_batch)
    p1.start()
    time.sleep(8)
    p2.start()



# from improveai.cythonized_feature_encoding import cfe
from improveai.decision import Decision
from improveai.decision_context import DecisionContext
from improveai.feature_encoder import FeatureEncoder
from improveai.decision_model import DecisionModel
from improveai.givens_provider import GivensProvider
from improveai.settings import USE_CYTHON_BACKEND

# import improveai.cythonized_feature_encoding.cythonized_feature_encoder as cfe
from improveai.cythonized_feature_encoding import cfe  # import FeatureEncoder as FastFeatureEncoder
FastFeatureEncoder = cfe.FeatureEncoder



#!python3
#cython: language_level=3

cdef extern from "npy_no_deprecated_api.h": pass

import cython
import numpy as np
cimport numpy as np


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np_1d_to_float32_array_inplace(np.ndarray converted_array):
    """
    Converts in-place numpy array to float32 dtype

    Parameters
    ----------
    converted_array: np.ndarray
        array to be converted to float32

    Returns
    -------

    """

    cdef np.ndarray f32_input_copy = converted_array.copy().astype(np.float32)

    # change dtype
    converted_array.dtype = np.float32
    # resize back to the original shape
    converted_array.resize((len(f32_input_copy), ), refcheck=False)
    # fill with original values
    converted_array[:] = f32_input_copy


cpdef encoded_variant_into_np_row(
        dict encoded_variant, list feature_names, np.ndarray into):
    """
    Fills in `into` row using provided variants and `feature_names`

    Parameters
    ----------
    encoded_variant: dict
        dict containing fully encoded variant (variant + givens)
    feature_names: np.ndarray or list
        collection of desired feature names
    into: np.ndarray
        array to be filled with encoded

    Returns
    -------
    None
        None

    """

    cdef dict hash_index_map = \
        {feature_hash: index for index, feature_hash
         in enumerate(feature_names)}

    cdef np.ndarray filler = \
        np.array(
            [(hash_index_map.get(feature_name, None), value)
             for feature_name, value in encoded_variant.items()
             if hash_index_map.get(feature_name, None) is not None])

    # sum into with encoded variants treating nans in sums as zeros

    cdef np.ndarray subset_index = np.empty(len(filler))

    if len(filler) > 0:

        subset_index = filler[:, 0].astype(int)

        into[subset_index] = np.nansum(
            np.array([into[subset_index], filler[:, 1]]), axis=0)

    # TODO wait until the conversion mechanism is determined
    # np_1d_to_float32_array_inplace(into)

@cython.boundscheck(False)
cpdef np.ndarray encode_variants_single_givens(
        list variants, dict givens, double noise,
        object variants_encoder, object givens_encoder):

    cdef dict into = {}

    cdef dict encoded_givens = \
        givens_encoder(givens, noise, into=dict(into))

    cdef int variants_count = len(variants)
    cdef np.ndarray res = np.empty((variants_count, ), dtype=object)

    cdef dict fully_encoded_variant = {}

    for variant_idx in range(variants_count):

        fully_encoded_variant = \
            variants_encoder(
                variants[variant_idx], noise, into=dict(encoded_givens))

        res[variant_idx] = fully_encoded_variant

    return res

@cython.boundscheck(False)
cpdef np.ndarray encode_variants_multiple_givens(
        list variants, list multiple_givens, list multiple_extra_features,
        double noise, object feature_encoder, object givens_encoder):

    assert len(variants) == len(multiple_givens)
    cdef int records_count = len(variants)
    cdef np.ndarray res = np.empty((records_count,), dtype=object)

    cdef dict empty_into = {}

    cdef dict givens = {}
    cdef dict extra_features = {}
    cdef dict encoded_givens = {}
    cdef dict fully_encoded_variant = {}

    for variant_idx in range(records_count):

        encoded_givens = {}
        givens = multiple_givens[variant_idx]
        if givens:
            encoded_givens = \
                givens_encoder(givens, noise, into=dict(empty_into))

        res[variant_idx] = \
            feature_encoder(variants[variant_idx], noise, into=encoded_givens)

        extra_features = multiple_extra_features[variant_idx]
        if extra_features:
            res[variant_idx].update(extra_features)

    return res


@cython.boundscheck(False)
cpdef double[:, :] encoded_variants_to_np(
        np.ndarray encoded_variants, list feature_names):

    cdef np.ndarray encoded_variants_array = \
        np.empty(shape=(len(encoded_variants), len(feature_names)), dtype=float)

    cdef dict name_to_index = \
        {feature_name: feature_index for feature_index, feature_name
         in enumerate(feature_names)}

    encoded_variants_array[:, :] = np.nan

    for encoded_variant_ixd, processed_variant in enumerate(encoded_variants):

        # this assumes that we have much less features per single variant that
        # total number of features
        filler = \
            np.array(
                [(name_to_index.get(feature_name, None), value)
                 for feature_name, value in processed_variant.items()
                 if name_to_index.get(feature_name, None) is not None])

        if len(filler) > 0:
            encoded_variants_array[encoded_variant_ixd, filler[:, 0]\
                .astype(int)] = filler[:, 1]

    return encoded_variants_array



import os
from setuptools import Extension, find_packages, setup

from Cython.Build import cythonize
import numpy as np

CYTHON_MODULE_DIR = 'cythonized_feature_encoding'
EXTENSION = 'pyx'
IMPROVE_DIR = 'improveai'


if __name__ == '__main__':

    cython_feature_encoding_utils_path_str = \
        os.sep.join(
            [IMPROVE_DIR, CYTHON_MODULE_DIR, 'cythonized_feature_encoding_utils.{}'.format(EXTENSION)])

    cython_feature_encoding_utils_ext = \
        Extension(
            '{}.{}.cythonized_feature_encoding_utils'.format(IMPROVE_DIR, CYTHON_MODULE_DIR),
            sources=[cython_feature_encoding_utils_path_str],
            include_dirs=[np.get_include(), os.sep.join(['.', IMPROVE_DIR, CYTHON_MODULE_DIR])])

    cython_feature_encoder_path_str = \
        os.sep.join(
            [IMPROVE_DIR, CYTHON_MODULE_DIR, 'cythonized_feature_encoder.{}'.format(EXTENSION)])

    cython_feature_encoder_ext = \
        Extension(
            '{}.{}.cythonized_feature_encoder'.format(IMPROVE_DIR, CYTHON_MODULE_DIR),
            sources=[cython_feature_encoder_path_str],
            include_dirs=[np.get_include(), os.sep.join(['.', IMPROVE_DIR, CYTHON_MODULE_DIR])])

    # TODO some requirements may depend on gcc version, e.g. coremltools
    install_requires = [
        "setuptools", "wheel",
        "Cython>=0.29.14",
        "xxhash==2.0.0",
        'coremltools==4.1',
        "requests>=2.24.0",
        "numpy>=1.19.2",
        "xgboost==1.4.2",
        "simplejson==3.17.2",
        "orjson",
        "svix-ksuid"]

    setup(
        name='improveai',
        version='7.0.1',
        description='Improve AI: AI Decisions for iOS, Android, and the Cloud',
        author='Justin Chapweske',
        author_email='support@improve.ai',
        url='https://github.com/improve-ai/python-sdk',
        packages=find_packages(exclude=['*.tox*', '*tests*']),
        install_requires=install_requires,
        ext_modules=cythonize([cython_feature_encoding_utils_ext, cython_feature_encoder_ext], language_level="3"),
        include_dirs=[np.get_include(), '.'],
        include_package_data=True,
        zip_safe=False)



from collections.abc import Iterable
import math
import numpy as np
import xxhash

from improveai import settings as improve_settings
from improveai.cythonized_feature_encoding import cfeu


xxhash3 = xxhash.xxh3_64_intdigest


class FeatureEncoder:
    """
    This class serves as a preprocessor which allows to
    convert event data in form of a JSON line (variants, givens and
    extra features) to the tabular format expected on input by XGBoost.

    Short summary of methods:
     - encode_givens() -> encodes givens / variant metadata to dict of
       feature name -> float pairs

     - encode_variant() -> encodes variant to dict of feature name -> float pairs

     - encode_feature_vector() -> encodes provided variant and givens;
       appends extra features to the encoded variant dict;
       converts encoded variant to numpy array using provided feature names

     - add_noise() -> sprinkles each value of the input dict with small noise

     - encode_to_np_matrix()  -> encodes collection of variants with collection
       of givens and extra features to the numpy 2D array / matrix. If desired
       uses cython backed (works much faster)

     - add_extra_features() - helper method for adding extra features to already
       encoded variants

    """

    VARIANT_METADATA_PARAMETER_NAME = 'givens'

    def __init__(self, model_seed):
        if (model_seed < 0):
            raise TypeError(
                "xxhash3_64 takes an unsigned 64-bit integer as seed. "
                "Seed must be greater than or equal to 0.")

        # memoize commonly used seeds
        self.variant_seed = xxhash3("variant", seed=model_seed)
        self.value_seed = xxhash3("$value", seed=self.variant_seed)
        self.givens_seed = xxhash3("givens", seed=model_seed)

    def _check_noise_value(self, noise: float):
        if noise > 1.0 or noise < 0.0:
            raise ValueError(
                'Provided `noise` is out of allowed bounds <0.0, 1.0>')

    def encode_givens(self, givens, noise=0.0, into=None):

        self._check_noise_value(noise=noise)

        if givens is not None and not isinstance(givens, dict):
            raise TypeError(
                "Only dict type is supported for context encoding. {} type was "
                "provided.".format(type(givens)))

        if into is None:
            into = {}

        encode(givens, self.givens_seed, shrink(noise), into)

        # TODO wait until the conversion mechanism is determined
        # self._convert_values_to_float32(into=into)
        return into

    def encode_variant(self, variant, noise=0.0, into=None):

        # TODO maybe this check is only a waste of runtime
        self._check_noise_value(noise=noise)

        if into is None:
            into = {}

        small_noise = shrink(noise)

        if isinstance(variant, dict):
            encode(variant, self.variant_seed, small_noise, into)
        else:
            encode(variant, self.value_seed, small_noise, into)

        # TODO wait until the conversion mechanism is determined
        # self._convert_values_to_float32(into=into)
        return into

    def encode_feature_vector(
            self, variant: dict = None, givens: dict = None,
            extra_features: dict = None, feature_names: list or np.ndarray = None,
            noise: float = 0.0, into: np.ndarray = None):

        if into is None:
            raise ValueError('`into` can`t be None')

        # encode givens
        encoded_givens = \
            self.encode_givens(givens=givens, noise=noise, into=dict())
        # encoded variant and givens
        encoded_variant = \
            self.encode_variant(
                variant=variant, noise=noise, into=encoded_givens)

        if extra_features:
            if not isinstance(extra_features, dict):
                raise TypeError(
                    'Provided `extra_features` is not a dict but: {}'
                    .format(extra_features))

            encoded_variant.update(extra_features)

        # n + nan = nan so you'll have to check for nan values on into
        # TODO once python-SDK is available from pip this should be changed
        if improve_settings.USE_CYTHON_BACKEND:
        # if USE_CYTHON_BACKEND:
            # i7 10th gen time per iter - 7.464-05 [s] / 0.96% of pure python time
            cfeu.encoded_variant_into_np_row(
                encoded_variant=encoded_variant, feature_names=feature_names, into=into)
        else:
            # i7 10th gen time per iter - 7.755e-05 [s]
            hash_index_map = \
                {feature_hash: index for index, feature_hash
                 in enumerate(feature_names)}

            filler = \
                np.array(
                    [(hash_index_map.get(feature_name, None), value)
                     for feature_name, value in encoded_variant.items()
                     if hash_index_map.get(feature_name, None) is not None])

            # sum into with encoded variants treating nans in sums as zeros
            if len(filler) > 0:
                subset_index = filler[:, 0].astype(int)

                into[subset_index] = np.nansum(
                    np.array([into[subset_index], filler[:, 1]]), axis=0)

        # TODO wait until the conversion mechanism is determined
        # np_to_float32_inplace(into)

    def encode_to_np_matrix(
            self, variants: Iterable, multiple_givens: Iterable,
            multiple_extra_features: Iterable, feature_names: Iterable,
            noise: float) -> np.ndarray:
        """
        Provided list of variants and corresponding lists of givens and extra
        features encodes variants completely and converts to numpy array

        Parameters
        ----------
        variants: list
            list of variants to be encoded
        multiple_givens: list
            list of givens - givens[i] will be used to encode variants[i]
        multiple_extra_features: list
            list of extra_features - multiple_extra_features[i] will be used
            to encode variants[i]
        feature_names: list
            names of features expected by the model
        noise: float
            noise to be used when encoding data

        Returns
        -------
        np.ndarray
            2D numpy array of encoded variants

        """

        if not isinstance(variants, list):
            variants = list(variants)

        if not isinstance(multiple_givens, list):
            multiple_givens = list(multiple_givens)

        if not isinstance(multiple_extra_features, list):
            multiple_extra_features = list(multiple_extra_features)

        if not isinstance(feature_names, list):
            feature_names = list(feature_names)

        # TODO once python-SDK is available from pip this should be changed
        if improve_settings.USE_CYTHON_BACKEND:
        # if USE_CYTHON_BACKEND:
            # i7 10th gen time per variant - 2.778e-05 [s] |
            # 33% of pure python implementation runtime per iter
            encoded_variants = cfeu.encode_variants_multiple_givens(
                variants=variants, multiple_givens=multiple_givens,
                multiple_extra_features=multiple_extra_features, noise=noise,
                feature_encoder=self.encode_variant,
                givens_encoder=self.encode_givens)
            encoded_variants_array = cfeu.encoded_variants_to_np(
                encoded_variants=encoded_variants, feature_names=feature_names)
        else:
            # i7 10th gen time per variant - 7.995e-05 [s]
            encoded_variants_array = np.full((len(variants), len(feature_names)), np.nan)

            [self.encode_feature_vector(
                variant=variant, givens=givens, extra_features=extra_features,
                feature_names=feature_names, noise=noise, into=into)
             for variant, givens, extra_features, into
             in zip(variants, multiple_givens, multiple_extra_features,
                    encoded_variants_array)]

        # TODO wait until the conversion mechanism is determined
        # return np.float32(encoded_variants_array)

        return encoded_variants_array

    def add_extra_features(self, encoded_variants: list, extra_features: list):
        """
        Once variants are encoded this method can be used to quickly append
        extra features

        Parameters
        ----------
        encoded_variants: list
            collection of encoded variants to be updated with extra features
        extra_features: list
            payload to be appended to encoded variants

        Returns
        -------
        None
            None

        """

        if extra_features is None:
            return

        if not isinstance(encoded_variants, list):
            raise TypeError('`encoded_variants` should be of a list type')

        if not isinstance(extra_features, list):
            raise TypeError('`extra_features` should be of a list type')

        [encoded_variant.update(single_extra_features)
         if single_extra_features is not None else encoded_variant
         for encoded_variant, single_extra_features in
         zip(encoded_variants, extra_features)]

    def _convert_values_to_float32(self, into: dict):
        """
        Converts all values in the input dict to float32

        Parameters
        ----------
        into: dict
            converted dict

        Returns
        -------

        """
        for k in into.keys():
            into[k] = np.float32(into[k])


# - None, json null, {}, [], nan, are treated as missing feature_names and ignored.
# NaN is technically not allowed in JSON but check anyway.
# - boolean true and false are encoded as 1.0 and 0.0 respectively.
# - strings are encoded into an additional one-hot feature.
# - numbers are encoded as-is with up to about 5 decimal places of precision
# - a small amount of noise is incorporated to avoid overfitting
# - feature names are 8 hexidecimal characters
def encode(object_, seed, small_noise, features):

    if isinstance(object_, (int, float)):  # bool is an instanceof int
        if math.isnan(object_):  # nan is treated as missing feature, return
            return

        feature_name = hash_to_feature_name(seed)

        previous_object_ = \
            _get_previous_value(
                feature_name=feature_name, into=features,
                small_noise=small_noise)

        features[feature_name] = sprinkle(object_ + previous_object_, small_noise)

        return

    if isinstance(object_, str):
        hashed = xxhash3(object_, seed=seed)

        feature_name = hash_to_feature_name(seed)

        previous_hashed = \
            _get_previous_value(
                feature_name=feature_name, into=features,
                small_noise=small_noise)

        features[feature_name] = \
            sprinkle(
                (((hashed & 0xffff0000) >> 16) - 0x8000) + previous_hashed,
                small_noise)

        hashed_feature_name = hash_to_feature_name(hashed)

        previous_hashed_for_feature_name = \
            _get_previous_value(
                feature_name=hashed_feature_name, into=features,
                small_noise=small_noise)

        features[hashed_feature_name] = \
            sprinkle(
                ((hashed & 0xffff) - 0x8000) + previous_hashed_for_feature_name,
                small_noise)

        return

    if isinstance(object_, dict):
        for key, value in object_.items():
            encode(
                value, xxhash3(key, seed=seed), small_noise, features)
        return

    if isinstance(object_, (list, tuple)):
        for index, item in enumerate(object_):
            encode(
                item, xxhash3(index.to_bytes(8, byteorder='big'), seed=seed),
                small_noise, features)
        return
    # None, json null, or unsupported type. Treat as missing feature, return


def _get_previous_value(
        feature_name: str, into: dict, small_noise: float):
    """
    Gets previous, 'unsprinkled' value of <feature_name> feature
    from <features> dict
    
    Parameters
    ----------
    feature_name: str
        name of a feature which previous value is desired
    into: dict
        dict containing current results of encoding
    small_noise: float
        small noise used to sprinkle

    Returns
    -------
    float
        'unsprinkled' value of desired feature

    """

    previous_sprinkled_object_ = into.get(feature_name, None)

    if previous_sprinkled_object_ is None:
        return 0.0
    else:
        return reverse_sprinkle(previous_sprinkled_object_, small_noise)


def hash_to_feature_name(hash_):
    return '%0*x' % (8, (hash_ >> 32))


def shrink(noise):
    return noise * 2 ** -17


def sprinkle(x, small_noise):
    return (x + small_noise) * (1 + small_noise)


def reverse_sprinkle(sprinkled_x, small_noise):
    """
    Retrieves true value from sprinkled one

    Parameters
    ----------
    sprinkled_x: float
        sprinkled value
    small_noise: float
        noise used to calculate sprinkled_Xx

    Returns
    -------
    float
        true value of x which was used to calculate sprinkled_x

    """
    return sprinkled_x / (1 + small_noise) - small_noise


# def add_noise(into, noise):
#     small_noise = shrink(noise)
#     for feature_name, value in into.items():
#         into[feature_name] = sprinkle(value, small_noise)


def np_to_float32_inplace(converted_array: np.ndarray):
    """
    Converts in-place numpy array to float32 dtype

    Parameters
    ----------
    converted_array: np.ndarray
        array to be converted to float32

    Returns
    -------

    """
    f32_input_copy = converted_array.copy().astype(np.float32)

    # change dtype
    converted_array.dtype = np.float32
    # resize back to the original shape
    converted_array.resize(f32_input_copy.shape, refcheck=False)
    # fill with original values
    converted_array[:] = f32_input_copy
